#
# Copyright (c) 2001 Bizar Software Pty Ltd (http://www.bizarsoftware.com.au/)
# This module is free software, and you may redistribute it and/or modify
# under the same terms as Python, so long as this copyright message and
# disclaimer are retained in their original form.
#
# IN NO EVENT SHALL BIZAR SOFTWARE PTY LTD BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
# OUT OF THE USE OF THIS CODE, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# BIZAR SOFTWARE PTY LTD SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE.  THE CODE PROVIDED HEREUNDER IS ON AN "AS IS"
# BASIS, AND THERE IS NO OBLIGATION WHATSOEVER TO PROVIDE MAINTENANCE,
# SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
# 
# $Id: dbinit.py,v 1.24 2003/09/22 02:03:34 jrouilla Exp $

import os

import config
from select_db import Database, Class, FileClass, IssueClass

# NYI - not yet implemented
# NYFI - not yet fully implemented

def open(name=None):
    ''' as from the roundupdb method openDB 
    ''' 
    from roundup.hyperdb import String, Password, Date, Link, Multilink
    from roundup.hyperdb import Interval, Boolean, Number

    # open the database
    db = Database(config, name)

    #
    # Now initialise the schema. Must do this each time the database is
    # opened.
    #

    # Class automatically gets these properties:
    #   creation = Date()
    #   activity = Date()
    #   creator = Link('user')
    # 
    # color is a property that was meant to provide a unique
    #    background/foreground color in issue lists. It's not
    #    currently implemented and may never be.
    # threshold is a property intended for automatic escallation of
    #    status using and external script. E.G. as the due date
    #    approaches, the new priority could be defined as:
    #    (threshold of current priority - threshold of max priority)
    #    divided by (number of days) - 1  till issue is due.
    #    This way issue would spend one day at highest priority.
    #    More complex escalation can be calculated using the lead time,
    #    Base priority determined from journal entries etc.
    # help is a property that describes when/where to use that priority.
    
    pri = Class(db, "priority", 
                    name=String(), order=String(), color=String,
                    threshold=Number(), help=String())
    pri.setkey("name")


    # help property lists a help message for the status.
    # abbreviation is a single character that is used to indicate
    #   status when listing issue numbers
    # transitions lists valid states to go to from that initial state.
    # requiredpermissions is a comma separated list of permissions
    #     that are required to transition to a given state. If the
    #     value is empty, no permission is required.
 
    stat = Class(db, "status", 
                    name=String(), order=String(), help=String(),
		    abbreviation=String(),
                    transitions=Multilink('status'),
		    requiredpermissions=String())
    stat.setkey("name")

    # help property lists a help message for the keyword.
    keyword = Class(db, "keyword", 
                    name=String(), help=String())
    keyword.setkey("name")

    # name property is name of type of message, "reply - to all" or
    #      "comment - to technical" are the defaults that describe the
    #      message type and where it is reported to.
    # fromaddr - meant to take the place of the the TRACKER_EMAIL
    #    and COMMENT_TRACKER_EMAIL for the detectors. Not sure it was
    #    a good idea. It can be a printf style format string that
    #    is used as a substitution template.
    # nosylists - nosy lists that get a copy of the message. Right now
    #    its hardcoded in the verynosy and nosy reactors.

    msgtype = Class(db, "msgtype",
                    name=String(),
                    fromaddr=String(),
                    nosylists=String(),
		    order=String())
    msgtype.setkey("name")

    query = Class(db, "query",
                    klass=String(),     name=String(),
                    url=String())
    query.setkey("name")


    # a queue is a filtering mechanism to steer an issue to a
    #    particular group of people. You can have queues by:
    #
    #    functional area: network, unix admin, Windows admin
    #        and specialists what only thie queues
    #
    #    administrative groups: test, training, integration,
    #                           development, customer support
    #        so admins that work for each group watch their own queue
    #
    #    Service level: level 1 support, level 2 support, level 3 support
    #        so second and third level support personnel only watch
    #        their queues.
    #
    #  The email property is used by newissuecopy to send notification
    #    emails when a new issue is created assigned to a given queue and
    #    nobody is assigned to the issue.
    #
    #    The email property will also be used by the queue reactor (not
    #    not yet written) to send email when an issue is moved from one
    #    queue to another if nobody is assigned to the issue.
    #    Multiple email addresses can be specified using "," between
    #    them. E.G. sysadmin1@example.com,sysadmin2@example.com
    #    This class is edited using the csv module, so the email string MUST
    #    be enclosed in quotes.

    queue = Class(db, "queue", 
                    name=String(),
                    email=String(),
		    order=String())
    queue.setkey("name")

    # add any additional database schema configuration here
    
    # Note: roles is a comma-separated string of Role names.
    # Timezone is a string like -5, +5 etc timezone offset
    # from GMT. Doesn't take into account daylight saving time etc.
    # Shouldn't it handle EST5EDT and things like that?
    user = Class(db, "user", 
                    username=String(),   password=Password(),
                    address=String(),    realname=String(), 
                    phone=String(),      organisation=String(),
                    alternate_addresses=String(),
                    queries=Multilink('query'), roles=String(),
                    timezone=String(),
                    nosy_keywords=Multilink('keyword'))

    user.setkey("username")

    # FileClass automatically gets these properties:
    #   content = String()    [saved to disk in <tracker home>/db/files/]
    #   (it also gets the Class properties creation, activity and creator)
    #   messagetype is used by nosy detector to determine which nosy
    #   list(s) to use.

    msg = FileClass(db, "msg", 
                    author=Link("user", do_journal='no'),
                    recipients=Multilink("user", do_journal='no'), 
                    date=Date(), summary=String(), 
                    files=Multilink("file"),
                    messageid=String(),  inreplyto=String(),
                    messagetype=Link("msgtype"),
                    cc=String(), times=Link("timelog"))


    # used for linking periods of time into messages and issues.
    timelog = Class(db, "timelog", period=Interval())


    # a class for storing uploded files.
    # status property allows a file to be set to obsolete or current NYI
    # acl property is intended to determine who can and can not view
    #     the file. NYI
    file = FileClass(db, "file",
                    name=String(), type=String(),
                    status=Link("filestatus"), acl = String())

    filestatus = Class(db, "filestatus",
                    name=String(), order=String())
    filestatus.setkey("name")

    # IssueClass automatically gets these properties:
    #   title = String()
    #   messages = Multilink("msg")
    #   files = Multilink("file")
    #   nosy = Multilink("user")
    #   superseder = Multilink("issue")
    #   (it also gets the Class properties creation, activity and creator)
    #
    # rouilj added:
    #
    #   startdate=Date() - date the task must be started, or changed
    #                      to open status. NYI
    #   duedate=Date() - date the task must be completed
    #   leadtime=Interval() - amount of time to complete the task
    #         use with due date to escallate priority
    #   timelog=Multilink("timelog") - used to record total time.
    #   requestedby=Link("user") - used to record person who requested
    #          this issue. It's auditor sets it to the creator of
    #          issue by default. It is useful for help desk purposes,
    #          where the person creating the issue is not the person
    #          requesting the issue.
    #   verynosy=Multilink("user") - used for nosy list for
    #        correspondance AND reply messages. Its the technical list
    #        in the web interface. FIXME change name to techlist
    #   fyi=String() - used to record summary notes and details
    #       that get buried in the ticket trail. Useful for noting
    #       ticket ID's for phone calls, serial numbers of equipment
    #       etc.
    #   queue = Link("queue"). Method of filtering ticket to particular
    #       people. See the queue class definition above.
    #   billinfo =String(). Store billing info for this ticket.
    #   needsreply=Boolean(). Used by the needsreply reactor.
    #             set to true if ticket is not new and the last person
    #             to add a message is not the assignedto person.
    #             If there is no assignedto person, then set to true
    #             if last message addition not done by person on
    #             technical (verynosy) list. This causes different
    #             display of issues that need to be responded to. It
    #             is useful if you are working in the field, and have
    #             a browser, but no email access to be notified of
    #             responses to issues. The stalled state can be
    #             used as well, but this is an additional hook.
    #   workingorder = Number(). Used to assign a working order to
    #               items. Useful when you have lots of things at the same
    #               priority.
    # interticket relationships:
    #   no code to enfornce these relationships has been written yet. NYI
    #   
    #   dependson relation if ticket A has ticket B listed in ticket
    #        A's dependson multilink property, that says that A can't
    #        be resolved until B is resolved.
    #   group relation if ticket A has ticket B and ticket C listed
    #        in ticket A's group it means that updates (messages,
    #        status changes etc) to ticket A should be applied to
    #        tickets B and C. You would use it when you have multiple
    #        tickets that are all solved by the same solution.
    #        Think of it as a merging of tickets that
    #        preserves the individuality of the tickets and permits
    #        unmerging if ticket C turns out to not be solved by the
    #        solution to ticket B.
    #   seealso - a general grouping solution. Since issues can be
    #        hyperlinked in the message bodies, this is somewhat
    #        redundant, but it is useful in that programs can easily
    #        find the links to publish a ticket and all related
    #        tickets without having to grovel through all of the messages.
    #
    #   Either the dependson or groups relation replaces the classic
    #        tracker "superseder" property whose meaning seems to be 
    #        incompletely defined.
    #
    #   The following implements a limited form of a schedulable action
    #   on a ticket. It allows setting an actiondate, actionstate and
    #   actionmessage that are removed after the actiondate has
    #   occurred. There is a cron script that must be run daily to
    #   actually perform the actions. It will not perform actions for
    #   resolved, dead, or delete tickets.
    #
    #   actiondate - a date whose time is ignored, only the date
    #                portion is used. On that day the action will occur.
    #   actionstatus - the issue will be changed to this status when the
    #                 action triggers. Used for unholding tickets.
    #   actioncomment - the comment to be sent when the action
    #                   occurs. It will be added to the issue.
    #                   Note that there is no actionreply.
    #   blockedby - issues blocking this issue from being resolved. It
    #               like the dependson attribute, but the dependson
    #               attribute is a permanent record. As items are resolved,
    #               they are removed from the blockedby list.
    # FIXME change topic to keywords
    issue = IssueClass(db, "issue", 
		    actionclearonmessage=Boolean(),
		    actioncomment=String(),
		    actiondate=Date(),
		    actionstatus=Link("status"),
                    assignedto=Link("user"),
                    billinfo=String(),
                    blockedby=Multilink("issue"),
                    dependson=Multilink("issue"),
                    duedate=Date(),
                    fyi=String(),
		    group=Multilink("issue"),
                    leadtime=Interval(),
		    needsreply=Boolean(),
                    priority=Link("priority"),
                    queue=Link("queue"),
                    requestedby=Link("user"),
                    seealso=Multilink("issue"),
                    startdate=Date(),
                    status=Link("status"),
                    timelog=Multilink("timelog"),
                    topic=Multilink("keyword"),
                    verynosy=Multilink("user"),
                    workingorder=Number()
                    )
    #
    # SECURITY SETTINGS
    #
    # See the configuration and customisation document for information
    # about security setup.
    # Add new Permissions for this schema
    for cl in 'issue', 'file', 'msg', 'user', 'query', 'keyword':
        db.security.addPermission(name="Edit", klass=cl,
            description="User is allowed to edit "+cl)
        db.security.addPermission(name="View", klass=cl,
            description="User is allowed to access "+cl)

    # Assign the access and edit Permissions for issue, file and message
    # to regular users now
    for cl in 'issue', 'file', 'msg', 'query', 'keyword':
        p = db.security.getPermission('View', cl)
        db.security.addPermissionToRole('User', p)
        p = db.security.getPermission('Edit', cl)
        db.security.addPermissionToRole('User', p)

    # and give the regular users access to the web and email interface
    p = db.security.getPermission('Web Access')
    db.security.addPermissionToRole('User', p)
    p = db.security.getPermission('Email Access')
    db.security.addPermissionToRole('User', p)

    # May users view other user information? Comment these lines out
    # if you don't want them to
    p = db.security.getPermission('View', 'user')
    db.security.addPermissionToRole('User', p)

    # Assign the appropriate permissions to the anonymous user's Anonymous
    # Role. Choices here are:
    # - Allow anonymous users to register through the web
    p = db.security.getPermission('Web Registration')
    db.security.addPermissionToRole('Anonymous', p)
    # - Allow anonymous (new) users to register through the email gateway
    p = db.security.getPermission('Email Registration')
    db.security.addPermissionToRole('Anonymous', p)
    # - Allow anonymous users access to the "issue" class of data
    #   Note: this also grants access to related information like files,
    #         messages, statuses etc that are linked to issues
    p = db.security.getPermission('View', 'issue')
    db.security.addPermissionToRole('Anonymous', p)
    # - Allow anonymous users access to edit the "issue" class of data
    #   Note: this also grants access to create related information like
    #         files and messages etc that are linked to issues
    #p = db.security.getPermission('Edit', 'issue')
    #db.security.addPermissionToRole('Anonymous', p)

    # oh, g'wan, let anonymous access the web interface too
    p = db.security.getPermission('Web Access')
    db.security.addPermissionToRole('Anonymous', p)

    # assign a role for a fixer
    db.security.addRole(name='Repair', description='Person who can fix problem')
    p = db.security.addPermission(name='Repair', klass='issue',
        description='User is allowed to be assigned to fix issues')
    p = db.security.getPermission('Repair', 'issue')

    db.security.addPermissionToRole('Repair', p)
    db.security.addPermissionToRole('Admin', p)

    # assign a role to allow setting to any state
    db.security.addRole(name='SetAnyState', description='User can set the state of an issue to any state')
    p = db.security.addPermission(name='SetAnyState', klass='issue',
        description='User can set the state of an issue to any state')
    p = db.security.getPermission('SetAnyState', 'issue')

    db.security.addPermissionToRole('SetAnyState', p)
    db.security.addPermissionToRole('Admin', p)
        
    import detectors
    detectors.init(db)

    # schema is set up - run any post-initialisation
    db.post_init()
    return db
 
def init(adminpw): 
    ''' as from the roundupdb method initDB 
 
    Open the new database, and add new nodes - used for initialisation. You
    can edit this before running the "roundup-admin initialise" command to
    change the initial database entries.
    ''' 
    dbdir = os.path.join(config.DATABASE, 'files')
    if not os.path.isdir(dbdir):
        os.makedirs(dbdir)

    db = open("admin")
    db.clear()

    #
    # INITIAL PRIORITY AND STATUS VALUES
    #
    pri = db.getclass('priority')
    pri.create(name="critical", order="1")
    pri.create(name="urgent", order="2")
    pri.create(name="high", order="3")
    pri.create(name="routine", order="4")
    pri.create(name="low", order="5")
    pri.create(name="minimum", order="6")

    stat = db.getclass('status')
    stat.create(name="new", order="1", abbreviation="N")
    stat.create(name="open", order="2", abbreviation="O")
    stat.create(name="stalled", order="3", abbreviation="S")
    stat.create(name="hold", order="4", abbreviation="H")
    stat.create(name="testing", order="5", abbreviation="T")
    stat.create(name="resolved", order="6", abbreviation="R")
    # C for copy of a ticket. Not quite the right semantics but...
    stat.create(name="duplicate", order="7", abbreviation="C")
    stat.create(name="dead", order="8", abbreviation="D")
    stat.create(name="delete", order="9", abbreviation="d")

    # create the two message types
    mtype = db.getclass('msgtype')
    mtype.create(name="comment - to technical",fromaddr=config.COMMENT_TRACKER_EMAIL,
         nosylists="verynosy", order="2")
    mtype.create(name="reply - to all", fromaddr=config.TRACKER_EMAIL,
         nosylists="nosy,verynosy", order="1")


    queue = db.getclass('queue')
    queue.create(name="Integration", email="donotsend", order="1")
    queue.create(name="Support", email="donotsend", order="2")
    queue.create(name="Test", email="donotsend", order="3")
    queue.create(name="Training", email="donotsend", order="4")
    queue.create(name="Other", email="donotsend", order="5")

    # define some default queries for:
    # showing watcher and technical people on tickets
    # showing all relationships between tickets dependson, groups, seealso
    # showing schedule relationship, due dates and lead time.
    query = db.getclass('query')
    query.create(klass="issue", name="WatcherTechnical",
                   url=":columns=title,id,status,queue,topic,duedate,creator,assignedto,activity,nosy,verynosy&:pagesize=50&:startwith=0")
    query.create(klass="issue",name="Relationships",
                   url=":columns=title,id,status,queue,duedate,assignedto,activity,dependson,group,seealso&:filter=status&status=-1,1,2,3,4,5&:pagesize=50&:startwith=0")
    query.create(klass="issue",name="Schedule",
                   url=":columns=title,id,status,queue,duedate,leadtime,startdate,creator,assignedto,activity&:sort=-duedate&:pagesize=50&:startwith=0&filter=status&status=-1,1,2,3,4,5&@sort=-leadtime&@group=duedate")

    # create the two default users
    user = db.getclass('user')
    user.create(username="admin", password=adminpw,
        address=config.ADMIN_EMAIL, roles='Admin')
    user.create(username="anonymous", roles='Anonymous')

    # used on file class.
    filestatus = db.getclass('filestatus')
    filestatus.create(name="current", order="1")
    filestatus.create(name="obsolete", order="2")

    # add any additional database create steps here - but only if you
    # haven't initialised the database with the admin "initialise" command

    db.commit()

# vim: set filetype=python ts=4 sw=4 et si

#SHA: 8168464dc6f1f68c1d8390bb954a1d84b00a4250
