#$Id: group_semantics.py,v 1.6 2003/05/11 07:18:04 jrouilla Exp $

def group_semantics(db, cl, nodeid, oldvalues):
    ''' Implement group semantics propigating changes to an issue

	This reactor implements the following semantics when invoked
	on the parent issue (the grouper) that groups a number of
	other tickets (the groupees). It propigates:

	  changes in grouper status to the groupees (unless the status
	     is new and the grouper was just created. This is
	     considered by me to be an error. It should at least have
	     status open.)

	  adds new messages to the grouper to the groupee's message
	  multilink

	This detector will will be invoked recursively since any
	changes it makes to the groupee will trigger this detector
	using the groupee as a new grouper.

	This detector is self terminating even if the groups are a
	loop. This termination is done by:

	  not changing the groupee's status if it is the same as the
	  grouper.

	and

	  not changing the message list (group multilink of issues) if
	    the grouper's new message has already been added to the
	    groupee's list.

        I have chosen to not propigate the assignedto parameter to
        provide a fast way of assigning multiple poeple to an issue.
	Simply create a new issue that groups only one issue. Changes
        to the grouper will propigate to the groupeee. Since multiple
        groupers can have the same groupee, this can be extended to
        multiple people.

	Should priority propigate, not necessarily. The group may
	have a higher priority than its groupee issues because
	it may be a routine priority for one person, but a high
	priority when 10 people are affected.

	Also queue and topics should not propigate.

	Arguably, setting the due date should trigger an auditor that
        verifies that all due dates of all groupees (which requires
        scanning the entire group relationship) occur later than the
        set due date.
    '''

    if __debug__:
        print "group_semantics.group_semantics[reactor]: triggered "
        print "group_semantics: nodeid, oldvalues", nodeid, oldvalues

    groupees = cl.get(nodeid, 'group')
    if __debug__:
        print "group_semantics: groupees for issue %s are: "%nodeid, groupees

    # ok so we have issues to propigate to
    # find out what we are propigating
    newmessages = determineNewMessages(cl, nodeid, oldvalues)

    if oldvalues is None:		   
       # this is a new issue so change children.
       status = cl.get(nodeid, 'status', cache=1)
       # if status is new, assume a mistake was made
       # and we are using the default new which shouldn't
       # be propigated lest we destroy useful status info.
       newstatus = db.status.lookup('new')
       if status is newstatus:
	  status = None
    elif oldvalues.has_key('status'):
       status = cl.get(nodeid, 'status', cache=1)
    else:
       # no status change so don't propigate anything.
       status = None

    if newmessages == [] and status is None:
        # there are no message or status changes.
	# just return
        return

    issueClass = cl.db.issue
    if newmessages != []:
        # propigate messages up and down the tree.
	# algorithm by Richard Jones
	#   determine any groupees which have me as grouper
	#     - ignore ALL of them if ANY have any of the messages already
	#   also determine each of my groupers
	#   for each message:
	#     for each groupee:
	#       check whether the message is in groupee's messages list
	#       if not, add it
	#     for each grouper:
	#       check whether the message is in groupers's messages list
	#       if not, add it
	# rouilj - this algorithm may fail to propagate messages in
        #          the case of a loop. I don't have a loop detection
	#          auditor in place yet.

        # scan all groupees for any of the messages.
	messagesPresent = False
	for issueid in groupees:
	    # propagate messages down the group tree
	    curmessages = issueClass.get(issueid, "messages")
	    for msg in newmessages:
		if msg in curmessages:
		   messagesPresent = True
		   if __debug__:
		      print "group_semantics: message ", msg, " in issue:", issueid
		   break
	    if messagesPresent:
	       break

	# if none of the new messages are present in my groupee's
	# assign the messages to the groupee's
	if not messagesPresent:
	    if __debug__:
	       print "group_semantics: add messages to groupees: ", groupees
  	    for issueid in groupees:
	        curmessages = issueClass.get(issueid, "messages")
		curmessages = curmessages + newmessages
		issueClass.set(issueid, messages=curmessages)
	else:
	    if __debug__:
		print "group_semantics: pruned all issues for ", groupees

	groupers = cl.find(group=nodeid)
	if __debug__:
		print "group_semantics: groupers are: ", groupers, "for node", nodeid
        for issueid in groupers:
	    curmessages = issueClass.get(issueid, "messages")
            messages_changed = False
            for msg in newmessages:
                if msg not in curmessages:
                   curmessages.append(msg)
                   messages_changed = True
	        else:
		    if __debug__:
		        print "group_semantics: node %s skipped issue "%nodeid, issueid, ", has message"

            if messages_changed:
                issueClass.set(issueid, messages=curmessages)
    else:
        if __debug__:
            print"group_semantics: No messages to propagate for issue", nodeid

    for issueid in groupees:
        # must do status last otherwise statusauditor will end up
	# reopening the ticket if the status is changed first and then
        # the messages are added.
	# Note: that this does not propagate upwards, only downwards.
        if status is not None:
            if issueClass.get(issueid, "status") is not status:
	        issueClass.set(issueid, status=status)

# loop detection may not be needed since we effectively
# leave a trail as we walk the nodes. Just in case though I'll
# leave this code here for future use
#    db.issue.react('set', detect_loop)
#    db.issue.react('create', detect_loop)

def detect_loop(db, cl, nodeid, newvalues):
    ''' Look for loops in the group settings.

	Walk all group properties making sure that no node is visited
	more than once. If we have a loop, report the nodes that point
	to it.
    '''
    seen={}
    walk_tree(db, cl, nodeid, nodeid, seen)
    
    return

def walk_tree(db, cl, groupee, grouper, seen):
    ''' Detect a loop by recursively walking the group tree
    '''
    # check to see if we have visited the groupee node before.
    # if so, raise an error, else record the visit.
    if seen.has_key(groupee):
       raise seen[groupee] + " has seen " + groupee
    else:
       seen[groupee] = grouper

    groupees = cl.get(groupee, 'group')
    
    for issueid in groupees:
	walk_tree(db, cl, issueid, groupee, seen)

# stolen from nosyreactor
def determineNewMessages(cl, nodeid, oldvalues):
    ''' Figure a list of the messages that are being added to the given
        node in this transaction.
    '''
    messages = []
    if oldvalues is None:
        # the action was a create, so use all the messages in the create
        messages = cl.get(nodeid, 'messages')
    elif oldvalues.has_key('messages'):
        # the action was a set (so adding new messages to an existing issue)
        m = {}
        for msgid in oldvalues['messages']:
            m[msgid] = 1
        messages = []
        # figure which of the messages now on the issue weren't there before
        for msgid in cl.get(nodeid, 'messages'):
            if not m.has_key(msgid):
                messages.append(msgid)
    return messages

def init(db):
    # fire after changes are made
    db.issue.react('set', group_semantics)
    db.issue.react('create', group_semantics)

# vim: set filetype=python ts=4 sw=4 et si
#SHA: 8b886dd30aaffe232873d1bcdeec179cdaa10aa2
